@@ -1,19 +1,32 @@
-import Cookie
+from ._compat import PY2, pickle, http_cookies, unicode_text, b64encode, b64decode, string_type
+
 import os
-from datetime import datetime, timedelta
 import time
-from beaker.crypto import hmac as HMAC, hmac_sha1 as SHA1, md5
+from datetime import datetime, timedelta
+from beaker.crypto import hmac as HMAC, hmac_sha1 as SHA1, sha1, get_nonce_size, DEFAULT_NONCE_BITS, get_crypto_module
 from beaker import crypto, util
 from beaker.cache import clsmap
 from beaker.exceptions import BeakerException, InvalidCryptoBackendError
-from base64 import b64encode, b64decode
+from beaker.cookie import SimpleCookie
+
+__all__ = ['SignedCookie', 'Session', 'InvalidSignature']
 
 
-__all__ = ['SignedCookie', 'Session']
+class _InvalidSignatureType(object):
+    """Returned from SignedCookie when the value's signature was invalid."""
+    def __nonzero__(self):
+        return False
+
+    def __bool__(self):
+        return False
+
+
+InvalidSignature = _InvalidSignatureType()
 
 
 try:
     import uuid
+
     def _session_id():
         return uuid.uuid4().hex
 except ImportError:
@@ -23,6 +36,7 @@ except ImportError:
     else:
         def getpid():
             return ''
+
     def _session_id():
         id_str = "%f%s%f%s" % (
                     time.time(),
@@ -30,66 +44,99 @@ except ImportError:
                     random.random(),
                     getpid()
                 )
-        if util.py3k:
-            return md5(
-                            md5(
-                                id_str.encode('ascii')
-                            ).hexdigest().encode('ascii')
-                        ).hexdigest()
+        # NB: nothing against second parameter to b64encode, but it seems
+        #     to be slower than simple chained replacement
+        if not PY2:
+            raw_id = b64encode(sha1(id_str.encode('ascii')).digest())
+            return str(raw_id.replace(b'+', b'-').replace(b'/', b'_').rstrip(b'='))
         else:
-            return md5(md5(id_str).hexdigest()).hexdigest()
+            raw_id = b64encode(sha1(id_str).digest())
+            return raw_id.replace('+', '-').replace('/', '_').rstrip('=')
 
-class SignedCookie(Cookie.BaseCookie):
+
+class SignedCookie(SimpleCookie):
     """Extends python cookie to give digital signature support"""
     def __init__(self, secret, input=None):
-        self.secret = secret
-        Cookie.BaseCookie.__init__(self, input)
+        self.secret = secret.encode('UTF-8')
+        http_cookies.BaseCookie.__init__(self, input)
 
     def value_decode(self, val):
         val = val.strip('"')
-        sig = HMAC.new(self.secret, val[40:], SHA1).hexdigest()
+        if not val:
+            return None, val
+
+        sig = HMAC.new(self.secret, val[40:].encode('utf-8'), SHA1).hexdigest()
 
         # Avoid timing attacks
         invalid_bits = 0
         input_sig = val[:40]
         if len(sig) != len(input_sig):
-            return None, val
+            return InvalidSignature, val
 
         for a, b in zip(sig, input_sig):
             invalid_bits += a != b
 
         if invalid_bits:
-            return None, val
+            return InvalidSignature, val
         else:
             return val[40:], val
 
     def value_encode(self, val):
-        sig = HMAC.new(self.secret, val, SHA1).hexdigest()
+        sig = HMAC.new(self.secret, val.encode('utf-8'), SHA1).hexdigest()
         return str(val), ("%s%s" % (sig, val))
 
 
 class Session(dict):
     """Session object that uses container package for storage.
 
-    ``key``
-        The name the cookie should be set to.
-    ``timeout``
-        How long session data is considered valid. This is used
-        regardless of the cookie being present or not to determine
-        whether session data is still valid.
-    ``cookie_domain``
-        Domain to use for the cookie.
-    ``secure``
-        Whether or not the cookie should only be sent over SSL.
-    ``httponly``
-        Whether or not the cookie should only be accessible by the browser
-        not by JavaScript.
+    :param invalidate_corrupt: How to handle corrupt data when loading. When
+                               set to True, then corrupt data will be silently
+                               invalidated and a new session created,
+                               otherwise invalid data will cause an exception.
+    :type invalidate_corrupt: bool
+    :param use_cookies: Whether or not cookies should be created. When set to
+                        False, it is assumed the user will handle storing the
+                        session on their own.
+    :type use_cookies: bool
+    :param type: What data backend type should be used to store the underlying
+                 session data
+    :param key: The name the cookie should be set to.
+    :param timeout: How long session data is considered valid. This is used
+                    regardless of the cookie being present or not to determine
+                    whether session data is still valid. Can be set to None to
+                    disable session time out.
+    :type timeout: int or None
+    :param save_accessed_time: Whether beaker should save the session's access
+                               time (True) or only modification time (False).
+                               Defaults to True.
+    :param cookie_expires: Expiration date for cookie
+    :param cookie_domain: Domain to use for the cookie.
+    :param cookie_path: Path to use for the cookie.
+    :param data_serializer: If ``"json"`` or ``"pickle"`` should be used
+                              to serialize data. Can also be an object with
+                              ``loads` and ``dumps`` methods. By default
+                              ``"pickle"`` is used.
+    :param secure: Whether or not the cookie should only be sent over SSL.
+    :param httponly: Whether or not the cookie should only be accessible by
+                     the browser not by JavaScript.
+    :param encrypt_key: The key to use for the local session encryption, if not
+                        provided the session will not be encrypted.
+    :param validate_key: The key used to sign the local encrypted session
+    :param encrypt_nonce_bits: Number of bits used to generate nonce for encryption key salt.
+                               For security reason this is 128bits be default. If you want
+                               to keep backward compatibility with sessions generated before 1.8.0
+                               set this to 48.
+    :param crypto_type: encryption module to use
     """
     def __init__(self, request, id=None, invalidate_corrupt=False,
                  use_cookies=True, type=None, data_dir=None,
-                 key='beaker.session.id', timeout=None, cookie_expires=True,
-                 cookie_domain=None, secret=None, secure=False,
-                 namespace_class=None, httponly=False, **namespace_args):
+                 key='beaker.session.id', timeout=None, save_accessed_time=True,
+                 cookie_expires=True, cookie_domain=None, cookie_path='/',
+                 data_serializer='pickle', secret=None,
+                 secure=False, namespace_class=None, httponly=False,
+                 encrypt_key=None, validate_key=None, encrypt_nonce_bits=DEFAULT_NONCE_BITS,
+                 crypto_type='default',
+                 **namespace_args):
         if not type:
             if data_dir:
                 self.type = 'file'
@@ -106,32 +153,52 @@ class Session(dict):
         self.data_dir = data_dir
         self.key = key
 
+        if timeout and not save_accessed_time:
+            raise BeakerException("timeout requires save_accessed_time")
         self.timeout = timeout
+        self.save_atime = save_accessed_time
         self.use_cookies = use_cookies
         self.cookie_expires = cookie_expires
 
+        self._set_serializer(data_serializer)
+
         # Default cookie domain/path
         self._domain = cookie_domain
-        self._path = '/'
+        self._path = cookie_path
         self.was_invalidated = False
         self.secret = secret
         self.secure = secure
         self.httponly = httponly
+        self.encrypt_key = encrypt_key
+        self.validate_key = validate_key
+        self.encrypt_nonce_size = get_nonce_size(encrypt_nonce_bits)
+        self.crypto_module = get_crypto_module(crypto_type)
         self.id = id
         self.accessed_dict = {}
+        self.invalidate_corrupt = invalidate_corrupt
 
         if self.use_cookies:
             cookieheader = request.get('cookie', '')
             if secret:
                 try:
-                    self.cookie = SignedCookie(secret, input=cookieheader)
-                except Cookie.CookieError:
-                    self.cookie = SignedCookie(secret, input=None)
+                    self.cookie = SignedCookie(
+                        secret,
+                        input=cookieheader,
+                    )
+                except http_cookies.CookieError:
+                    self.cookie = SignedCookie(
+                        secret,
+                        input=None,
+                    )
             else:
-                self.cookie = Cookie.SimpleCookie(input=cookieheader)
+                self.cookie = SimpleCookie(input=cookieheader)
 
             if not self.id and self.key in self.cookie:
-                self.id = self.cookie[self.key].value
+                cookie_data = self.cookie[self.key].value
+                # Should we check invalidate_corrupt here?
+                if cookie_data is InvalidSignature:
+                    cookie_data = None
+                self.id = cookie_data
 
         self.is_new = self.id is None
         if self.is_new:
@@ -140,8 +207,8 @@ class Session(dict):
         else:
             try:
                 self.load()
-            except Exception, e:
-                if invalidate_corrupt:
+            except Exception as e:
+                if self.invalidate_corrupt:
                     util.warn(
                         "Invalidating corrupt session %s; "
                         "error was: %s.  Set invalidate_corrupt=False "
@@ -150,6 +217,20 @@ class Session(dict):
                 else:
                     raise
 
+    def _set_serializer(self, data_serializer):
+        self.data_serializer = data_serializer
+        if self.data_serializer == 'json':
+            self.serializer = util.JsonSerializer()
+        elif self.data_serializer == 'pickle':
+            self.serializer = util.PickleSerializer()
+        elif isinstance(self.data_serializer, string_type):
+            raise BeakerException('Invalid value for data_serializer: %s' % data_serializer)
+        else:
+            self.serializer = data_serializer
+
+    def has_key(self, name):
+        return name in self
+
     def _set_cookie_values(self, expires=None):
         self.cookie[self.key] = self.id
         if self._domain:
@@ -163,24 +244,28 @@ class Session(dict):
 
     def _set_cookie_expires(self, expires):
         if expires is None:
-            if self.cookie_expires is not True:
-                if self.cookie_expires is False:
-                    expires = datetime.fromtimestamp( 0x7FFFFFFF )
-                elif isinstance(self.cookie_expires, timedelta):
-                    expires = datetime.utcnow() + self.cookie_expires
-                elif isinstance(self.cookie_expires, datetime):
-                    expires = self.cookie_expires
-                else:
-                    raise ValueError("Invalid argument for cookie_expires: %s"
-                                     % repr(self.cookie_expires))
-            else:
-                expires = None
-        if expires is not None:
-            self.cookie[self.key]['expires'] = \
-                expires.strftime("%a, %d-%b-%Y %H:%M:%S GMT" )
-        return expires
+            expires = self.cookie_expires
+        if expires is False:
+            expires_date = datetime.fromtimestamp(0x7FFFFFFF)
+        elif isinstance(expires, timedelta):
+            expires_date = datetime.utcnow() + expires
+        elif isinstance(expires, datetime):
+            expires_date = expires
+        elif expires is not True:
+            raise ValueError("Invalid argument for cookie_expires: %s"
+                             % repr(self.cookie_expires))
+        self.cookie_expires = expires
+        if not self.cookie or self.key not in self.cookie:
+            self.cookie[self.key] = self.id
+        if expires is True:
+            self.cookie[self.key]['expires'] = ''
+            return True
+        self.cookie[self.key]['expires'] = \
+            expires_date.strftime("%a, %d-%b-%Y %H:%M:%S GMT")
+        return expires_date
 
     def _update_cookie_out(self, set_cookie=True):
+        self._set_cookie_values()
         self.request['cookie_out'] = self.cookie[self.key].output(header='')
         self.request['set_cookie'] = set_cookie
 
@@ -188,7 +273,7 @@ class Session(dict):
         try:
             if self.httponly:
                 self.cookie[self.key]['httponly'] = True
-        except Cookie.CookieError, e:
+        except http_cookies.CookieError as e:
             if 'Invalid Attribute httponly' not in str(e):
                 raise
             util.warn('Python 2.6+ is required to use httponly')
@@ -200,8 +285,7 @@ class Session(dict):
             self.is_new = True
             self.last_accessed = None
         if self.use_cookies:
-            self._set_cookie_values()
-            sc = set_new == False
+            sc = set_new is False
             self._update_cookie_out(set_cookie=sc)
 
     @property
@@ -209,8 +293,7 @@ class Session(dict):
         return self['_creation_time']
 
     def _set_domain(self, domain):
-        self['_domain'] = domain
-        self.cookie[self.key]['domain'] = domain
+        self['_domain'] = self._domain = domain
         self._update_cookie_out()
 
     def _get_domain(self):
@@ -220,7 +303,6 @@ class Session(dict):
 
     def _set_path(self, path):
         self['_path'] = self._path = path
-        self.cookie[self.key]['path'] = path
         self._update_cookie_out()
 
     def _get_path(self):
@@ -228,9 +310,42 @@ class Session(dict):
 
     path = property(_get_path, _set_path)
 
+    def _encrypt_data(self, session_data=None):
+        """Serialize, encipher, and base64 the session dict"""
+        session_data = session_data or self.copy()
+        if self.encrypt_key:
+            nonce_len, nonce_b64len = self.encrypt_nonce_size
+            nonce = b64encode(os.urandom(nonce_len))[:nonce_b64len]
+            encrypt_key = crypto.generateCryptoKeys(self.encrypt_key,
+                                                    self.validate_key + nonce,
+                                                    1,
+                                                    self.crypto_module.getKeyLength())
+            data = self.serializer.dumps(session_data)
+            return nonce + b64encode(self.crypto_module.aesEncrypt(data, encrypt_key))
+        else:
+            data = self.serializer.dumps(session_data)
+            return b64encode(data)
+
+    def _decrypt_data(self, session_data):
+        """Base64, decipher, then un-serialize the data for the session
+        dict"""
+        if self.encrypt_key:
+            __, nonce_b64len = self.encrypt_nonce_size
+            nonce = session_data[:nonce_b64len]
+            encrypt_key = crypto.generateCryptoKeys(self.encrypt_key,
+                                                    self.validate_key + nonce,
+                                                    1,
+                                                    self.crypto_module.getKeyLength())
+            payload = b64decode(session_data[nonce_b64len:])
+            data = self.crypto_module.aesDecrypt(payload, encrypt_key)
+        else:
+            data = b64decode(session_data)
+
+        return self.serializer.loads(data)
+
     def _delete_cookie(self):
         self.request['set_cookie'] = True
-        expires = datetime.utcnow().replace(year=2003)
+        expires = datetime.utcnow() - timedelta(365)
         self._set_cookie_values(expires)
         self._update_cookie_out()
 
@@ -266,24 +381,34 @@ class Session(dict):
             try:
                 session_data = self.namespace['session']
 
+                if (session_data is not None and self.encrypt_key):
+                    session_data = self._decrypt_data(session_data)
+
                 # Memcached always returns a key, its None when its not
                 # present
                 if session_data is None:
                     session_data = {
-                        '_creation_time':now,
-                        '_accessed_time':now
+                        '_creation_time': now,
+                        '_accessed_time': now
                     }
                     self.is_new = True
             except (KeyError, TypeError):
                 session_data = {
-                    '_creation_time':now,
-                    '_accessed_time':now
+                    '_creation_time': now,
+                    '_accessed_time': now
+                }
+                self.is_new = True
+
+            if session_data is None or len(session_data) == 0:
+                session_data = {
+                    '_creation_time': now,
+                    '_accessed_time': now
                 }
                 self.is_new = True
 
             if self.timeout is not None and \
                now - session_data['_accessed_time'] > self.timeout:
-                timed_out= True
+                timed_out = True
             else:
                 # Properly set the last_accessed time, which is different
                 # than the *currently* _accessed_time
@@ -294,7 +419,7 @@ class Session(dict):
 
                 # Update the current _accessed_time
                 session_data['_accessed_time'] = now
-                
+
                 # Set the path if applicable
                 if '_path' in session_data:
                     self._path = session_data['_path']
@@ -315,7 +440,7 @@ class Session(dict):
         """
         # Look to see if its a new session that was only accessed
         # Don't save it under that case
-        if accessed_only and self.is_new:
+        if accessed_only and (self.is_new or not self.save_atime):
             return None
 
         # this session might not have a namespace yet or the session id
@@ -334,6 +459,9 @@ class Session(dict):
             else:
                 data = dict(self.items())
 
+            if self.encrypt_key:
+                data = self._encrypt_data(data)
+
             # Save the data
             if not data and 'session' in self.namespace:
                 del self.namespace['session']
@@ -355,7 +483,7 @@ class Session(dict):
             creates a new session id, retains all session data
 
             Its a good security practice to regnerate the id after a client
-            elevates priviliges.
+            elevates privileges.
 
         """
         self._create_id(set_new=False)
@@ -385,51 +513,70 @@ class Session(dict):
         """
         self.namespace.release_write_lock()
 
+
 class CookieSession(Session):
     """Pure cookie-based session
 
     Options recognized when using cookie-based sessions are slightly
     more restricted than general sessions.
 
-    ``key``
-        The name the cookie should be set to.
-    ``timeout``
-        How long session data is considered valid. This is used
-        regardless of the cookie being present or not to determine
-        whether session data is still valid.
-    ``encrypt_key``
-        The key to use for the session encryption, if not provided the
-        session will not be encrypted.
-    ``validate_key``
-        The key used to sign the encrypted session
-    ``cookie_domain``
-        Domain to use for the cookie.
-    ``secure``
-        Whether or not the cookie should only be sent over SSL.
-    ``httponly``
-        Whether or not the cookie should only be accessible by the browser
-        not by JavaScript.
-
+    :param key: The name the cookie should be set to.
+    :param timeout: How long session data is considered valid. This is used
+                    regardless of the cookie being present or not to determine
+                    whether session data is still valid.
+    :type timeout: int
+    :param save_accessed_time: Whether beaker should save the session's access
+                               time (True) or only modification time (False).
+                               Defaults to True.
+    :param cookie_expires: Expiration date for cookie
+    :param cookie_domain: Domain to use for the cookie.
+    :param cookie_path: Path to use for the cookie.
+    :param data_serializer: If ``"json"`` or ``"pickle"`` should be used
+                              to serialize data. Can also be an object with
+                              ``loads` and ``dumps`` methods. By default
+                              ``"pickle"`` is used.
+    :param secure: Whether or not the cookie should only be sent over SSL.
+    :param httponly: Whether or not the cookie should only be accessible by
+                     the browser not by JavaScript.
+    :param encrypt_key: The key to use for the local session encryption, if not
+                        provided the session will not be encrypted.
+    :param validate_key: The key used to sign the local encrypted session
+    :param invalidate_corrupt: How to handle corrupt data when loading. When
+                               set to True, then corrupt data will be silently
+                               invalidated and a new session created,
+                               otherwise invalid data will cause an exception.
+    :type invalidate_corrupt: bool
+    :param crypto_type: The crypto module to use.
     """
     def __init__(self, request, key='beaker.session.id', timeout=None,
-                 cookie_expires=True, cookie_domain=None, encrypt_key=None,
-                 validate_key=None, secure=False, httponly=False, **kwargs):
+                 save_accessed_time=True, cookie_expires=True, cookie_domain=None,
+                 cookie_path='/', encrypt_key=None, validate_key=None, secure=False,
+                 httponly=False, data_serializer='pickle',
+                 encrypt_nonce_bits=DEFAULT_NONCE_BITS, invalidate_corrupt=False,
+                 crypto_type='default',
+                 **kwargs):
 
-        if not crypto.has_aes and encrypt_key:
+        self.crypto_module = get_crypto_module(crypto_type)
+
+        if not self.crypto_module.has_aes and encrypt_key:
             raise InvalidCryptoBackendError("No AES library is installed, can't generate "
-                                  "encrypted cookie-only Session.")
+                                            "encrypted cookie-only Session.")
 
         self.request = request
         self.key = key
         self.timeout = timeout
+        self.save_atime = save_accessed_time
         self.cookie_expires = cookie_expires
         self.encrypt_key = encrypt_key
         self.validate_key = validate_key
+        self.encrypt_nonce_size = get_nonce_size(encrypt_nonce_bits)
         self.request['set_cookie'] = False
         self.secure = secure
         self.httponly = httponly
         self._domain = cookie_domain
-        self._path = '/'
+        self._path = cookie_path
+        self.invalidate_corrupt = invalidate_corrupt
+        self._set_serializer(data_serializer)
 
         try:
             cookieheader = request['cookie']
@@ -439,11 +586,19 @@ class CookieSession(Session):
         if validate_key is None:
             raise BeakerException("No validate_key specified for Cookie only "
                                   "Session.")
+        if timeout and not save_accessed_time:
+            raise BeakerException("timeout requires save_accessed_time")
 
         try:
-            self.cookie = SignedCookie(validate_key, input=cookieheader)
-        except Cookie.CookieError:
-            self.cookie = SignedCookie(validate_key, input=None)
+            self.cookie = SignedCookie(
+                validate_key,
+                input=cookieheader,
+            )
+        except http_cookies.CookieError:
+            self.cookie = SignedCookie(
+                validate_key,
+                input=None,
+            )
 
         self['_id'] = _session_id()
         self.is_new = True
@@ -452,13 +607,27 @@ class CookieSession(Session):
         if self.key in self.cookie and self.cookie[self.key].value is not None:
             self.is_new = False
             try:
-                self.update(self._decrypt_data())
+                cookie_data = self.cookie[self.key].value
+                if cookie_data is InvalidSignature:
+                    raise BeakerException("Invalid signature")
+                self.update(self._decrypt_data(cookie_data))
                 self._path = self.get('_path', '/')
-            except:
-                pass
-            if self.timeout is not None and time.time() - \
-               self['_accessed_time'] > self.timeout:
-                self.clear()
+            except Exception as e:
+                if self.invalidate_corrupt:
+                    util.warn(
+                        "Invalidating corrupt session %s; "
+                        "error was: %s.  Set invalidate_corrupt=False "
+                        "to propagate this exception." % (self.id, e))
+                    self.invalidate()
+                else:
+                    raise
+
+            if self.timeout is not None:
+                now = time.time()
+                last_accessed_time = self.get('_accessed_time', now)
+                if now - last_accessed_time > self.timeout:
+                    self.clear()
+
             self.accessed_dict = self.copy()
             self._create_cookie()
 
@@ -487,35 +656,9 @@ class CookieSession(Session):
 
     path = property(_get_path, _set_path)
 
-    def _encrypt_data(self):
-        """Serialize, encipher, and base64 the session dict"""
-        if self.encrypt_key:
-            nonce = b64encode(os.urandom(40))[:8]
-            encrypt_key = crypto.generateCryptoKeys(self.encrypt_key,
-                                             self.validate_key + nonce, 1)
-            data = util.pickle.dumps(self.copy(), 2)
-            return nonce + b64encode(crypto.aesEncrypt(data, encrypt_key))
-        else:
-            data = util.pickle.dumps(self.copy(), 2)
-            return b64encode(data)
-
-    def _decrypt_data(self):
-        """Bas64, decipher, then un-serialize the data for the session
-        dict"""
-        if self.encrypt_key:
-            nonce = self.cookie[self.key].value[:8]
-            encrypt_key = crypto.generateCryptoKeys(self.encrypt_key,
-                                             self.validate_key + nonce, 1)
-            payload = b64decode(self.cookie[self.key].value[8:])
-            data = crypto.aesDecrypt(payload, encrypt_key)
-            return util.pickle.loads(data)
-        else:
-            data = b64decode(self.cookie[self.key].value)
-            return util.pickle.loads(data)
-
     def save(self, accessed_only=False):
         """Saves the data for this session to persistent storage"""
-        if accessed_only and self.is_new:
+        if accessed_only and (self.is_new or not self.save_atime):
             return
         if accessed_only:
             self.clear()
@@ -534,6 +677,12 @@ class CookieSession(Session):
             self['_id'] = _session_id()
         self['_accessed_time'] = time.time()
 
+        val = self._encrypt_data()
+        if len(val) > 4064:
+            raise BeakerException("Cookie value is too long to store")
+
+        self.cookie[self.key] = val
+
         if '_expires' in self:
             expires = self['_expires']
         else:
@@ -542,11 +691,6 @@ class CookieSession(Session):
         if expires is not None:
             self['_expires'] = expires
 
-        val = self._encrypt_data()
-        if len(val) > 4064:
-            raise BeakerException("Cookie value is too long to store")
-
-        self.cookie[self.key] = val
         if '_domain' in self:
             self.cookie[self.key]['domain'] = self['_domain']
         elif self._domain:
@@ -568,7 +712,7 @@ class CookieSession(Session):
 
     def invalidate(self):
         """Clear the contents and start a new session"""
-        self.delete()
+        self.clear()
         self['_id'] = _session_id()
 
 
@@ -585,20 +729,25 @@ class SessionObject(object):
         self.__dict__['_params'] = params
         self.__dict__['_environ'] = environ
         self.__dict__['_sess'] = None
-        self.__dict__['_headers'] = []
+        self.__dict__['_headers'] = {}
 
     def _session(self):
         """Lazy initial creation of session object"""
         if self.__dict__['_sess'] is None:
             params = self.__dict__['_params']
             environ = self.__dict__['_environ']
-            self.__dict__['_headers'] = req = {'cookie_out':None}
+            self.__dict__['_headers'] = req = {'cookie_out': None}
             req['cookie'] = environ.get('HTTP_COOKIE')
-            if params.get('type') == 'cookie':
-                self.__dict__['_sess'] = CookieSession(req, **params)
+            session_cls = params.get('session_class', None)
+            if session_cls is None:
+                if params.get('type') == 'cookie':
+                    session_cls = CookieSession
+                else:
+                    session_cls = Session
             else:
-                self.__dict__['_sess'] = Session(req, use_cookies=True,
-                                                 **params)
+                assert issubclass(session_cls, Session),\
+                    "Not a Session: " + session_cls
+            self.__dict__['_sess'] = session_cls(req, **params)
         return self.__dict__['_sess']
 
     def __getattr__(self, attr):
@@ -627,7 +776,10 @@ class SessionObject(object):
         return iter(self._session().keys())
 
     def __contains__(self, key):
-        return self._session().has_key(key)
+        return key in self._session()
+
+    def has_key(self, key):
+        return key in self._session()
 
     def get_by_id(self, id):
         """Loads a session given a session ID"""
@@ -647,21 +799,28 @@ class SessionObject(object):
     def persist(self):
         """Persist the session to the storage
 
-        If its set to autosave, then the entire session will be saved
-        regardless of if save() has been called. Otherwise, just the
-        accessed time will be updated if save() was not called, or
-        the session will be saved if save() was called.
+        Always saves the whole session if save() or delete() have been called.
+        If they haven't:
+
+        - If autosave is set to true, saves the the entire session regardless.
+        - If save_accessed_time is set to true or unset, only saves the updated
+          access time.
+        - If save_accessed_time is set to false, doesn't save anything.
 
         """
         if self.__dict__['_params'].get('auto'):
             self._session().save()
-        else:
-            if self.__dict__.get('_dirty'):
+        elif self.__dict__['_params'].get('save_accessed_time', True):
+            if self.dirty():
                 self._session().save()
             else:
                 self._session().save(accessed_only=True)
+        else:  # save_accessed_time is false
+            if self.dirty():
+                self._session().save()
 
     def dirty(self):
+        """Returns True if save() or delete() have been called"""
         return self.__dict__.get('_dirty', False)
 
     def accessed(self):
